#include "icca.h"
#include "launcher/launcher.h"
#include "rawinput/rawinput.h"
#include "games/popn/io.h"
#include "misc/eamuse.h"

using namespace GameAPI;

// settings
namespace acio {
    bool ICCA_NUMPAD_ENABLE = true;
    bool ICCA_TOPROW_ENABLE = true;
    bool ICCA_FLIP_ROWS = false;
}

/*
 * Helpers
 */

struct ICCA_STATUS {
    char status_code;
    char solenoid;
    char front_sensor;
    char rear_sensor;
    char uid[8];
    int error;
    unsigned int key_edge;
    unsigned int key_level;
};
enum ICCA_WORKFLOW {
    STEP,
    SLEEP,
    START,
    INIT,
    READY,
    GET_USERID,
    ACTIVE,
    EJECT,
    EJECT_CHECK,
    END,
    CLOSE_EJECT,
    CLOSE_E_CHK,
    CLOSE_END,
    ERR_GETUID = -2
};
struct ICCA_UNIT {
    struct ICCA_STATUS status = {};
    enum ICCA_WORKFLOW state = STEP;
    bool card_cmd_pressed = false;
    bool card_in = false;
    char key_serial = 0;
    bool uid_skip = false;
    bool initialized = false;
};
static struct ICCA_UNIT ICCA_UNITS[2];

static inline int icca_get_active_count() {
    int active_count = 0;
    for (auto unit : ICCA_UNITS)
        active_count += unit.initialized;
    return active_count;
}

static inline bool icca_is_active(int unit_id) {
    if (icca_get_active_count() == 1)
        if (unit_id < 2)
            return true;
    return (GetAsyncKeyState(VK_NUMLOCK) > 0) == ((unit_id & 1) > 0);
}

static inline int icca_get_unit_id(int unit_id) {
    if (icca_get_active_count() < 2)
        return 1;
    else {
        if (unit_id > 1)
            return 1;
        else
            return 0;
    }
}

static inline void update_card(int unit_id) {
    static bool kb_insert_press_old[2] = {false, false};
    bool kb_insert_press = false;

    // naive keypress
    if (eamuse_keypad_state_naive()) {
        kb_insert_press = icca_is_active(unit_id) && (
                (acio::ICCA_TOPROW_ENABLE && GetAsyncKeyState(VK_BACK))
                || (acio::ICCA_NUMPAD_ENABLE && GetAsyncKeyState(VK_ADD)));
        if (!kb_insert_press) {
            kb_insert_press_old[0] = false;
            kb_insert_press_old[1] = false;
        }
    }

    // eamio keypress
    int index = unit_id > 0 && icca_get_active_count() > 1 ? 1 : 0;
    kb_insert_press |= eamuse_get_keypad_state(index) & (1 << EAM_IO_INSERT);

    // get unit
    ICCA_UNIT *unit = &ICCA_UNITS[unit_id];

    // check for card insert
    if (eamuse_card_insert_consume(icca_get_active_count(), unit_id) ||
        (kb_insert_press && !kb_insert_press_old[unit_id])) {
        if (!unit->card_cmd_pressed) {
            unit->card_cmd_pressed = true;
            if (unit->state == GET_USERID || unit->state == CLOSE_EJECT || unit->state == STEP) {
                if (unit->uid_skip || eamuse_get_card(icca_get_active_count(), unit_id, unit->status.uid)) {
                    unit->state = ACTIVE;
                    unit->status.error = 0;
                } else {
                    memset(unit->status.uid, 0, 8);
                    unit->state = ERR_GETUID;
                }
                unit->card_in = true;
            } else if (unit->state == EJECT_CHECK) {
                unit->state = SLEEP;
                unit->card_in = false;
            }
        } else {
            unit->state = ACTIVE;
            unit->status.error = 0;
            unit->card_in = true;
        }
    } else {
        unit->card_cmd_pressed = false;
        unit->state = CLOSE_EJECT;
    }

    // save state
    kb_insert_press_old[unit_id] = kb_insert_press;
}

static bool KEYPAD_LAST[2][12];
static int KEYPAD_KEYS[] = {
        VK_NUMPAD0,
        VK_NUMPAD1,
        VK_NUMPAD2,
        VK_NUMPAD3,
        VK_NUMPAD4,
        VK_NUMPAD5,
        VK_NUMPAD6,
        VK_NUMPAD7,
        VK_NUMPAD8,
        VK_NUMPAD9,
        VK_RETURN, // enter
        VK_OEM_COMMA // double zero
};
static int KEYPAD_KEYS_ALT[] = {
        '0',
        '1',
        '2',
        '3',
        '4',
        '5',
        '6',
        '7',
        '8',
        '9',
        VK_BACK, // enter
        VK_OEM_PERIOD // double zero
};
static int KEYPAD_EAMUSE_MAPPING[] = {
        0, 1, 5, 9, 2, 6, 10, 3, 7, 11, 8, 4
};
static int KEYPAD_KEY_CODES[]{
        0x100,
        0x200,
        0x2000,
        2,
        0x400,
        0x4000,
        4,
        0x800,
        0x8000,
        8,
        1,
        0x1000
};
static int KEYPAD_KEY_CODE_NUMS[]{
        0, 1, 5, 9, 2, 6, 10, 3, 7, 11, 8, 4
};

static inline void keypad_update(int unit_id) {

    // check for active unit
    bool active = icca_is_active(unit_id) && eamuse_keypad_state_naive();

    // reset unit
    struct ICCA_UNIT *unit = &ICCA_UNITS[unit_id];
    unit->status.key_edge = 0;
    unit->status.key_level = 0;

    // get eamu state
    int index = unit_id > 0 && icca_get_active_count() > 1 ? 1 : 0;
    uint16_t eamu_state = eamuse_get_keypad_state(index);

    // iterate keypad
    for (int n = 0; n < 12; n++) {
        int i = n;

        // flip 123 with 789
        if (acio::ICCA_FLIP_ROWS) {
            if (!n)
                i = 11;
            else if (n < 4)
                i = n + 6;
            else if (n > 6 && n < 10)
                i = n - 6;
            else if (n == 11)
                i = 0;
        }

        // check if pressed
        if ((active && acio::ICCA_NUMPAD_ENABLE && GetAsyncKeyState(KEYPAD_KEYS[i])) ||
            (active && acio::ICCA_TOPROW_ENABLE && GetAsyncKeyState(KEYPAD_KEYS_ALT[i])) ||
            (eamu_state & (1 << KEYPAD_EAMUSE_MAPPING[i]))) {
            unit->status.key_level |= KEYPAD_KEY_CODES[i];
            if (!KEYPAD_LAST[unit_id][i]) {
                unit->status.key_edge = (unsigned int) KEYPAD_KEY_CODES[n] << 16;
                unit->status.key_edge |= unit->key_serial << 4;
                unit->status.key_edge |= 0x80 | KEYPAD_KEY_CODE_NUMS[n];
                unit->key_serial = (char) (unit->key_serial + 1) & (char) 0x07;
            }
            KEYPAD_LAST[unit_id][i] = true;
        } else
            KEYPAD_LAST[unit_id][i] = false;
    }
}

/*
 * Implementations
 */

static bool __cdecl ac_io_icca_cardunit_init(int unit_id) {
    unit_id = icca_get_unit_id(unit_id);

    // dirty workaround code
    if (icca_get_active_count() < 1)
        ICCA_UNITS[unit_id].initialized = true;
    else {
        ICCA_UNITS[0].initialized = true;
        ICCA_UNITS[1].initialized = true;
    }

    // initial poll
    eamuse_get_keypad_state(unit_id);

    // return success
    return true;
}

static char __cdecl ac_io_icca_cardunit_init_isfinished(int unit_id, DWORD *status) {
    unit_id = icca_get_unit_id(unit_id);
    *status = READY;
    return 1;
}

static char __cdecl ac_io_icca_crypt_init(int unit_id) {
    unit_id = icca_get_unit_id(unit_id);
    return 1;
}

static char __cdecl ac_io_icca_device_control_iccard_power_supply_off(int unit_id) {
    unit_id = icca_get_unit_id(unit_id);
    return 1;
}

static char __cdecl ac_io_icca_device_control_iccard_power_supply_on(int unit_id) {
    unit_id = icca_get_unit_id(unit_id);
    return 1;
}

static bool __cdecl ac_io_icca_device_control_isfinished(int unit_id, DWORD *a2) {
    unit_id = icca_get_unit_id(unit_id);
    return true;
}

static bool __cdecl ac_io_icca_get_keep_alive_error(int unit_id, DWORD *a2) {
    unit_id = icca_get_unit_id(unit_id);
    *a2 = 0;
    return false;
}

static char __cdecl ac_io_icca_get_status(void *status, void *unit_id) {
    if ((long long) unit_id > 2) {
        void *swap = status;
        status = unit_id;
        unit_id = swap;
    }
    unit_id = (void *) ((long long) icca_get_unit_id((int) ((long long) unit_id)));
    keypad_update((int) ((long long) unit_id));
    update_card((int) ((long long) unit_id));
    ICCA_UNITS[(long long) unit_id].status.status_code = ICCA_UNITS[(long long) unit_id].state;
    memcpy(status, &ICCA_UNITS[(long long) unit_id].status, sizeof(struct ICCA_STATUS));
    return 1;
}

static char __cdecl ac_io_icca_get_uid(int unit_id, char *card) {
    unit_id = icca_get_unit_id(unit_id);
    ICCA_UNIT *unit = &ICCA_UNITS[unit_id];

    // copy card
    memcpy(card, unit->status.uid, 8);

    // check for error
    return unit->state != ERR_GETUID;
}

static char __cdecl ac_io_icca_get_uid_felica(int unit_id, char *card) {
    unit_id = icca_get_unit_id(unit_id);
    ICCA_UNIT *unit = &ICCA_UNITS[unit_id];

    // copy card
    memcpy(card, unit->status.uid, 8);

    // set felica flag
    bool felica = unit->status.uid[0] != (char) 0xE0 && unit->status.uid[1] != (char) 0x04;
    card[8] = (char) (felica ? 1 : 0);

    // check for error
    return unit->state != ERR_GETUID;
}

static bool __cdecl ac_io_icca_is_felica() {
    return true;
}

static char __cdecl ac_io_icca_req_uid(int unit_id) {
    unit_id = icca_get_unit_id(unit_id);
    ICCA_UNIT *unit = &ICCA_UNITS[unit_id];
    unit->state = GET_USERID;
    update_card(unit_id);
    return 1;
}

static int __cdecl ac_io_icca_req_uid_isfinished(int unit_id, DWORD *read_state) {
    unit_id = icca_get_unit_id(unit_id);
    ICCA_UNIT *unit = &ICCA_UNITS[unit_id];
    if (unit->card_in) {
        unit->state = unit->card_in ? END : CLOSE_EJECT;
        unit->card_in = false;
    }
    *read_state = unit->state;
    return 1;
}

static int __cdecl ac_io_icca_send_keep_alive_packet(int a1, int a2, int a3) {
    return 0;
}

static int __cdecl ac_io_icca_workflow(int workflow, int unit_id) {
    unit_id = icca_get_unit_id(unit_id);
    ICCA_UNIT *unit = &ICCA_UNITS[unit_id];
    switch (workflow) {
        case STEP:
            unit->state = STEP;
            break;
        case SLEEP:
            unit->state = SLEEP;
            break;
        case INIT:
            unit->state = READY;
            break;
        case START:
            if (unit->card_in)
                unit->state = ACTIVE;
            else
                unit->state = READY;
            break;
        case EJECT:
            unit->card_in = false;
            break;
        case CLOSE_EJECT:
            unit->state = unit->card_in ? EJECT_CHECK : SLEEP;
            break;
        case CLOSE_END:
            unit->state = SLEEP;
            break;
        case GET_USERID:
            unit->state = GET_USERID;
            break;
        default:
            break;
    }

    return unit->state;
}

static char __cdecl ac_io_icca_req_status(int a1, char a2) {
    return 1;
}

static bool __cdecl ac_io_icca_req_status_isfinished(int a1, int *a2) {
    *a2 = 11;
    return true;
}

/*
 * Module stuff
 */

acio::ICCAModule::ICCAModule(HMODULE module, acio::HookMode hookMode) : ACIOModule("ICCA", module, hookMode) {
}

void acio::ICCAModule::attach() {
    ACIOModule::attach();

    // hooks
    ACIO_MODULE_HOOK(ac_io_icca_cardunit_init);
    ACIO_MODULE_HOOK(ac_io_icca_cardunit_init_isfinished);
    ACIO_MODULE_HOOK(ac_io_icca_crypt_init);
    ACIO_MODULE_HOOK(ac_io_icca_device_control_iccard_power_supply_off);
    ACIO_MODULE_HOOK(ac_io_icca_device_control_iccard_power_supply_on);
    ACIO_MODULE_HOOK(ac_io_icca_device_control_isfinished);
    ACIO_MODULE_HOOK(ac_io_icca_get_keep_alive_error);
    ACIO_MODULE_HOOK(ac_io_icca_get_status);
    ACIO_MODULE_HOOK(ac_io_icca_get_uid);
    ACIO_MODULE_HOOK(ac_io_icca_get_uid_felica);
    ACIO_MODULE_HOOK(ac_io_icca_is_felica);
    ACIO_MODULE_HOOK(ac_io_icca_req_uid);
    ACIO_MODULE_HOOK(ac_io_icca_req_uid_isfinished);
    ACIO_MODULE_HOOK(ac_io_icca_send_keep_alive_packet);
    ACIO_MODULE_HOOK(ac_io_icca_workflow);
    ACIO_MODULE_HOOK(ac_io_icca_req_status);
    ACIO_MODULE_HOOK(ac_io_icca_req_status_isfinished);
}
