#ifndef SPICETOOLS_RAWINPUT_H
#define SPICETOOLS_RAWINPUT_H

#include <vector>
#include <thread>
#include <windows.h>
#include "device.h"
#include "hotplug.h"

namespace rawinput {

    struct DeviceCallback {
        void *data;
        std::function<void(void*, Device*)> f;
    };

    class RawInputManager {
    private:

        HotplugManager* hotplug;
        std::vector<Device> devices;
        HWND input_hwnd = nullptr;
        WNDCLASSEX input_hwnd_class {};
        std::thread* input_thread = nullptr;
        std::thread* flush_thread = nullptr;
        bool flush_thread_running = false;
        std::vector<DeviceCallback> callback_add;
        std::vector<DeviceCallback> callback_change;

        void input_hwnd_create();
        void input_hwnd_destroy();
        void devices_reload();
        void devices_scan_rawinput(RAWINPUTDEVICELIST *device, bool log=true);
        void devices_scan_midi();
        void devices_scan_piuio();
        void devices_destruct();
        void devices_destruct(Device *device, bool log=true);
        void flush_start();
        void flush_stop();

        static std::string rawinput_get_device_name(HANDLE hDevice);
        static std::string rawinput_get_device_description(DeviceInfo info, std::string device_name);

        static LRESULT CALLBACK input_wnd_proc(HWND, UINT, WPARAM, LPARAM);
        static void CALLBACK input_midi_proc(HMIDIIN, UINT, DWORD_PTR, DWORD_PTR, DWORD_PTR);
        static DeviceInfo get_device_info(std::string device_name);

    public:

        RawInputManager();
        ~RawInputManager();

        void sextet_register(std::string port_name, std::string alias="Sextet", bool warn=true);
        void devices_scan_rawinput(std::string device_name="");
        void devices_remove(std::string name);

        void devices_register();
        void devices_unregister();

        static void device_write_output(Device* device, bool only_updated = true);
        void devices_flush_output(bool only_updated = true);

        __stdcall void devices_print();
        Device* devices_get(std::string &name, bool updated=false);

        inline std::vector<Device>* devices_get() {
            return &devices;
        }

        inline std::vector<Device*> devices_get_updated() {
            std::vector<Device*> updated;
            for (auto &device : *devices_get()) {
                device.mutex->lock();
                if (device.updated) {
                    device.updated = false;
                    device.mutex->unlock();
                    updated.emplace_back(&device);
                } else
                    device.mutex->unlock();
            }
            return updated;
        }

        inline void devices_midi_freeze(bool freeze) {
            for (auto &device : *devices_get()) {
                device.mutex->lock();
                if (device.type == MIDI) {
                    device.midiInfo->freeze = freeze;
                    if (!freeze)
                        for (unsigned short index = 0; index < 16 * 128; index++)
                            device.midiInfo->states[index] = false;
                }
                device.mutex->unlock();
            }
        }

        inline void add_callback_add(void *data, std::function<void(void*, Device*)> callback) {
            this->callback_add.push_back(DeviceCallback {
                .data = data,
                .f = callback,
            });
        }

        inline void add_callback_change(void *data, std::function<void(void*, Device*)> callback) {
            this->callback_change.push_back(DeviceCallback {
                    .data = data,
                    .f = callback,
            });
        }
    };
}

#endif //SPICETOOLS_RAWINPUT_H
